// Copyright (c) 2003-2010 Christopher M. Kohlhoff
// Modifications (c) 2011 by Guilherme Maciel Ferreira / Intra2net AG
//
// Distributed under the Boost Software License, Version 1.0.
//    (See accompanying file LICENSE_1_0.txt or copy at
//          http://www.boost.org/LICENSE_1_0.txt)
#include "icmp/icmpv6header.h"

#include "boost_assert_handler.h"
#include "icmp/icmpdestinationunreachablemessage.h"
#include "icmp/icmpechoreplymessage.h"
#include "icmp/icmpechorequestmessage.h"
#include "icmp/icmpgenericmessage.h"

using namespace std;
using boost::shared_ptr;

//-----------------------------------------------------------------------------
// Icmpv6Header
//-----------------------------------------------------------------------------

static const size_t HeaderSizeInBytes = 4;

Icmpv6Header::Icmpv6Header() :
    MessageFormat()
{
}

Icmpv6Header::Icmpv6Header(
        const Icmpv6Type type,
        const uint8_t code,
        const uint16_t checksum,
        const uint16_t identifier,
        const uint16_t sequence_number
) :
    MessageFormat()
{
    set_icmp_message_format( type );

    set_type( type );
    set_code( code );
    set_checksum( checksum );
    set_identifier( identifier );
    set_sequence_number( sequence_number );
}

Icmpv6Type Icmpv6Header::get_type() const
{
    return get_icmp_message_format()->get_type_v6();
}

void Icmpv6Header::set_type( const Icmpv6Type type )
{
    get_icmp_message_format()->set_type_v6( type );
} //lint !e1762

uint8_t Icmpv6Header::get_code() const
{
    return get_icmp_message_format()->get_code();
}

void Icmpv6Header::set_code( const uint8_t code )
{
    get_icmp_message_format()->set_code( code );
} //lint !e1762

uint16_t Icmpv6Header::get_checksum() const
{
    return get_icmp_message_format()->get_checksum();
}

void Icmpv6Header::set_checksum( const uint16_t checksum )
{
    get_icmp_message_format()->set_checksum( checksum );
} //lint !e1762

uint16_t Icmpv6Header::get_identifier() const
{
    return get_icmp_message_format()->get_identifier();
}

void Icmpv6Header::set_identifier( const uint16_t identifier )
{
    get_icmp_message_format()->set_identifier( identifier );
} //lint !e1762

uint16_t Icmpv6Header::get_sequence_number() const
{
    return get_icmp_message_format()->get_sequence_number();
}

void Icmpv6Header::set_sequence_number( const uint16_t sequence_number )
{
    get_icmp_message_format()->set_sequence_number( sequence_number );
} //lint !e1762

shared_ptr<IcmpMessage> Icmpv6Header::get_icmp_message_format() const
{
    BOOST_ASSERT( MessageFormat.get() != NULL );

    return MessageFormat;
}

void Icmpv6Header::set_icmp_message_format( const Icmpv6Type type )
{
    BOOST_ASSERT( MessageFormat.get() == NULL );

    if ( MessageFormat.get() == NULL )
    {
        switch ( type )
        {
            case Icmpv6Type_EchoReply:
                MessageFormat.reset( new IcmpEchoReplyMessage );
                break;
            case Icmpv6Type_EchoRequest:
                MessageFormat.reset( new IcmpEchoRequestMessage );
                break;
            case Icmpv6Type_DestinationUnreachable:
                MessageFormat.reset(
                        new IcmpDestinationUnreachableMessage
                );
                break;
            case Icmpv6Type_PacketTooBig:
            case Icmpv6Type_TimeExceeded:
            case Icmpv6Type_ParameterProblem:
            case Icmpv6Type_RouterSolicitation:
            case Icmpv6Type_RouterAdvertisement:
            case Icmpv6Type_NeighborSolicitation:
            case Icmpv6Type_NeighborAdvertisement:
            case Icmpv6Type_RedirectMessage:
            case Icmpv6Type_RouterRenumbering:
            case Icmpv6Type_ICMPNodeInformationQuery:
            case Icmpv6Type_ICMPNodeInformationResponse:
            case Icmpv6Type_InverseNeighborDiscoverySolicitationMessage:
            case Icmpv6Type_InverseNeighborDiscoveryAdvertisementMessage:
            case Icmpv6Type_MulticastListenerDiscovery:
            case Icmpv6Type_HomeAgentAddressDiscoveryRequestMessage:
            case Icmpv6Type_HomeAgentAddressDiscoveryReplyMessage:
            case Icmpv6Type_MobilePrefixSolicitation:
            case Icmpv6Type_MobilePrefixAdvertisement:
            case Icmpv6Type_CertificationPathSolicitation:
            case Icmpv6Type_CertificationPathAdvertisement:
            case Icmpv6Type_MulticastRouterAdvertisement:
            case Icmpv6Type_MulticastRouterSolicitation:
            case Icmpv6Type_MulticastRouterTermination:
            case Icmpv6Type_Generic:
                MessageFormat.reset( new IcmpGenericMessage );
                break;
            case Icmpv6Type_First:
            case Icmpv6Type_InvalidLast:
            default:
                BOOST_ASSERT( !"Try to set an invalid ICMPv6 message type" );
                break;
        }

        MessageFormat->init( type );
    }

    BOOST_ASSERT( MessageFormat.get() != NULL );
    BOOST_ASSERT( MessageFormat->get_type_v6() != Icmpv6Type_InvalidLast );
}

void Icmpv6Header::set_icmp_message_format( std::istream &is )
{
    // read the first byte, which contains the type of the ICMP message
    uint8_t first_byte = 0;
    (void) is.read( reinterpret_cast<char *>(&first_byte), 1 );

    // must keep the stream intact, so place the read byte back
    (void) is.putback( static_cast<std::istream::char_type>(first_byte) );

    // now select the message format for the given type
    Icmpv6Type header_type = static_cast<Icmpv6Type>( first_byte );
    set_icmp_message_format( header_type );
}

size_t Icmpv6Header::get_header_length() const
{
    return HeaderSizeInBytes;
}

std::istream& operator>>(
        std::istream &is,
        Icmpv6Header &header
)
{
    // select the message format which will parse the fields
    header.set_icmp_message_format( is );

    // delegate the parsing of the fields to a specialized decoder
    return header.get_icmp_message_format()->read( is );
}

std::ostream& operator<<(
        std::ostream &os,
        const Icmpv6Header &header
)
{
    // delegate the writing of the fields to a specialized encoder
    return header.get_icmp_message_format()->write( os );
}
